@@ -1,4 +1,10 @@
-// For license information, see the COPYING file in the qconf base directory.
+/*
+Copyright (C) 2004-2008  Justin Karneges
+
+This file is free software; unlimited permission is given to copy and/or
+distribute it, with or without modifications, as long as this notice is
+preserved.
+*/
 
 #include "conf4.h"
 
@@ -25,7 +31,16 @@ QStringList qc_pathlist()
 	QStringList list;
 	QString path = qc_getenv("PATH");
 	if(!path.isEmpty())
+	{
+#ifdef Q_OS_WIN
+		list = path.split(';', QString::SkipEmptyParts);
+#else
 		list = path.split(':', QString::SkipEmptyParts);
+#endif
+	}
+#ifdef Q_OS_WIN
+	list.prepend(".");
+#endif
 	return list;
 }
 
@@ -41,61 +56,93 @@ QString qc_findprogram(const QString &prog)
 			out = fi.filePath();
 			break;
 		}
+
+#ifdef Q_OS_WIN
+		// on windows, be sure to look for .exe
+		if(prog.right(4).toLower() != ".exe")
+		{
+			fi = QFileInfo(list[n] + '/' + prog + ".exe");
+			if(fi.exists() && fi.isExecutable())
+			{
+				out = fi.filePath();
+				break;
+			}
+		}
+#endif
 	}
 	return out;
 }
 
 QString qc_findself(const QString &argv0)
 {
+#ifdef Q_OS_WIN
+	if(argv0.contains('\\\\'))
+#else
 	if(argv0.contains('/'))
+#endif
 		return argv0;
 	else
 		return qc_findprogram(argv0);
 }
 
-int qc_runcommand(const QString &command, QByteArray *out, bool showOutput)
+int qc_run_program_or_command(const QString &prog, const QStringList &args, const QString &command, QByteArray *out, bool showOutput)
 {
-	QString fullcmd = command;
-	if(!showOutput)
-		fullcmd += " 2>/dev/null";
-	FILE *f = popen(fullcmd.toLatin1().data(), "r");
-	if(!f)
-		return -1;
 	if(out)
 		out->clear();
-	while(1)
-	{
-		char c = (char)fgetc(f);
-		if(feof(f))
-			break;
-		if(out)
-			out->append(c);
-		if(showOutput)
-			fputc(c, stdout);
-	}
-	int ret = pclose(f);
-	if(ret == -1)
+
+	QProcess process;
+	process.setReadChannel(QProcess::StandardOutput);
+
+	if(!prog.isEmpty())
+		process.start(prog, args);
+	else if(!command.isEmpty())
+		process.start(command);
+	else
 		return -1;
-	return ret;
+
+	if(!process.waitForStarted(-1))
+		return -1;
+
+	QByteArray buf;
+
+	while(process.waitForReadyRead(-1))
+	{
+		buf = process.readAllStandardOutput();
+		if(out)
+			out->append(buf);
+		if(showOutput)
+			fprintf(stdout, "%s", buf.data());
+
+		buf = process.readAllStandardError();
+		if(showOutput)
+			fprintf(stderr, "%s", buf.data());
+	}
+
+	buf = process.readAllStandardError();
+	if(showOutput)
+		fprintf(stderr, "%s", buf.data());
+
+	// calling waitForReadyRead will cause the process to eventually be
+	//   marked as finished, so we should not need to separately call
+	//   waitForFinished. however, we will do it anyway just to be safe.
+	//   we won't check the return value since false could still mean
+	//   success (if the process had already been marked as finished).
+	process.waitForFinished(-1);
+
+	if(process.exitStatus() != QProcess::NormalExit)
+		return -1;
+
+	return process.exitCode();
+}
+
+int qc_runcommand(const QString &command, QByteArray *out, bool showOutput)
+{
+	return qc_run_program_or_command(QString(), QStringList(), command, out, showOutput);
 }
 
 int qc_runprogram(const QString &prog, const QStringList &args, QByteArray *out, bool showOutput)
 {
-	QString fullcmd = prog;
-	QString argstr = args.join(" ");
-	if(!argstr.isEmpty())
-		fullcmd += QString(" ") + argstr;
-	return qc_runcommand(fullcmd, out, showOutput);
-
-	// TODO: use QProcess once it is fixed
-	/*
-	QProcess process;
-	if(showOutput)
-		process.setReadChannelMode(ForwardedChannels);
-	process.start(prog, args);
-	process.waitForFinished(-1);
-	return process.exitCode();
-	*/
+	return qc_run_program_or_command(prog, args, QString(), out, showOutput);
 }
 
 bool qc_removedir(const QString &dirPath)
@@ -360,7 +407,12 @@ int Conf::doCommand(const QString &prog, const QStringList &args, QByteArray *ou
 
 bool Conf::doCompileAndLink(const QString &filedata, const QStringList &incs, const QString &libs, const QString &proextra, int *retcode)
 {
+#ifdef Q_OS_WIN
+	QDir tmp("qconftemp");
+#else
 	QDir tmp(".qconftemp");
+#endif
+
 	if(!tmp.mkdir("atest"))
 	{
 		debug("unable to create atest dir");
@@ -393,6 +445,7 @@ bool Conf::doCompileAndLink(const QString &filedata, const QStringList &incs, co
 	QString pro = QString(
 		"CONFIG  += console\n"
 		"CONFIG  -= qt app_bundle\n"
+		"DESTDIR  = $$PWD\n"
 		"SOURCES += atest.cpp\n");
 	QString inc = incs.join(" ");
 	if(!inc.isEmpty())
@@ -429,7 +482,13 @@ bool Conf::doCompileAndLink(const QString &filedata, const QStringList &incs, co
 		{
 			ok = true;
 			if(retcode)
-				*retcode = doCommand(QString("./") + out, QStringList());
+			{
+				QString runatest = out;
+#ifdef Q_OS_UNIX
+				runatest.prepend("./");
+#endif
+				*retcode = doCommand(runatest, QStringList());
+			}
 		}
 		r = doCommand(maketool, QStringList() << "distclean");
 		if(r != 0)
@@ -706,8 +765,8 @@ void Conf::addExtra(const QString &str)
 int main()
 {
 	Conf *conf = new Conf;
-	ConfObj *o;
-	o = 0;
+	ConfObj *o = 0;
+	Q_UNUSED(o);
 #ifdef HAVE_MODULES
 # include"modules_new.cpp"
 #endif
@@ -721,6 +780,7 @@ int main()
 	QString confCommand = qc_getenv("QC_COMMAND");
 	QString proName = qc_getenv("QC_PROFILE");
 	conf->qmake_path = qc_getenv("QC_QMAKE");
+	conf->qmakespec = qc_getenv("QC_QMAKESPEC");
 	conf->maketool = qc_getenv("QC_MAKETOOL");
 
 	if(conf->debug_enabled)
@@ -744,6 +804,7 @@ int main()
 		printf("builddir:     [%s]\n", qPrintable(builddir));
 		printf("profile:      [%s]\n", qPrintable(proPath));
 		printf("qmake path:   [%s]\n", qPrintable(conf->qmake_path));
+		printf("qmakespec:    [%s]\n", qPrintable(conf->qmakespec));
 		printf("make tool:    [%s]\n", qPrintable(conf->maketool));
 		printf("\n");
 	}
@@ -801,13 +862,21 @@ int main()
 		success = true;
 	}
 	QString qmake_path = conf->qmake_path;
+	QString qmakespec = conf->qmakespec;
 	delete conf;
 
 	if(!success)
 		return 1;
 
 	// run qmake on the project file
-	int ret = qc_runprogram(qmake_path, QStringList() << proPath, 0, true);
+	QStringList args;
+	if(!qmakespec.isEmpty())
+	{
+		args += "-spec";
+		args += qmakespec;
+	}
+	args += proPath;
+	int ret = qc_runprogram(qmake_path, args, 0, true);
 	if(ret != 0)
 		return 1;
 
