@@ -1,6 +1,6 @@
 /*
  * qconf.cpp - main qconf source
- * Copyright (C) 2003-2007  Justin Karneges
+ * Copyright (C) 2003-2009  Justin Karneges
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -26,7 +26,7 @@
 
 #include "stringhelp.h"
 
-#define VERSION "1.4"
+#define VERSION "1.5"
 
 #define CONF_USAGE_SPACE 4
 #define CONF_WRAP        78
@@ -106,6 +106,93 @@ QString echoBlock(int tabs, const QString &in)
 	return str;
 }
 
+QString formatBlock(const QString &in)
+{
+	QStringList pars = in.split('\n', QString::KeepEmptyParts);
+	if(!pars.isEmpty() && pars.last().isEmpty())
+		pars.removeLast();
+	QStringList lines;
+	int n;
+	for(n = 0; n < pars.count(); ++n) {
+		QStringList list = wrapString(pars[n], CONF_WRAP);
+		for(int n2 = 0; n2 < list.count(); ++n2)
+			lines.append(list[n2]);
+	}
+
+	QString str;
+	for(n = 0; n < lines.count(); ++n)
+		str += lines[n] + '\n';
+	return str;
+}
+
+static void write32(quint8 *out, quint32 i)
+{
+	out[0] = (i >> 24) & 0xff;
+	out[1] = (i >> 16) & 0xff;
+	out[2] = (i >> 8) & 0xff;
+	out[3] = i & 0xff;
+}
+
+static QByteArray lenval(const QByteArray &in)
+{
+	QByteArray out;
+	out.resize(4);
+	write32((quint8 *)out.data(), in.size());
+	out += in;
+	return out;
+}
+
+static QByteArray embed_file(const QString &name, const QByteArray &data)
+{
+	QByteArray out;
+	out += lenval(name.toLatin1());
+	out += lenval(data);
+	return out;
+}
+
+// eat double-backslashes on windows, as we don't need the extra escaping to
+//   satisfy the unix shell script
+static QByteArray filter_backslashes(const QByteArray &in)
+{
+	QByteArray out;
+
+	bool inbs = false;
+	for(int n = 0; n < in.size(); ++n)
+	{
+		if(in[n] == '\\')
+		{
+			if(inbs)
+			{
+				inbs = false;
+				continue;
+			}
+
+			inbs = true;
+		}
+		else
+			inbs = false;
+
+		out += in[n];
+	}
+
+	return out;
+}
+
+static QByteArray get_configexe_stub()
+{
+	QFile f;
+	f.setFileName(":/configexe/configexe_stub.exe");
+	if(!f.open(QIODevice::ReadOnly))
+	{
+		// for debugging purposes, if .exe stub isn't found, use
+		//   possible unix stub
+		f.setFileName(":/configexe/configexe_stub");
+		if(!f.open(QIODevice::ReadOnly))
+			return QByteArray();
+	}
+	return f.readAll();
+}
+
 const char *qt4_info_str =
 	"Be sure you have a proper Qt 4.0 build environment set up.  This means not just Qt, "
 	"but also a C++ compiler, a make tool, and any other packages necessary "
@@ -125,6 +212,22 @@ const char *qt4_info_str =
 	"recommended options.  #1 and #2 are mainly for overriding the system configuration.\n"
 	"\n";
 
+const char *qt4_info_str_win =
+	"Be sure you have a proper Qt 4.0 build environment set up.  This means not just Qt, "
+	"but also a C++ compiler, a make tool, and any other packages necessary "
+	"for compiling C++ programs.\n"
+	"\n"
+	"If you are certain everything is installed, then it could be that Qt 4 is not being "
+	"recognized or that a different version of Qt is being detected by mistake (for example, "
+	"this could happen if %QTDIR% is pointing to a Qt 3 installation).  At least one of "
+	"the following conditions must be satisfied:\n"
+	"\n"
+	" 1) --qtdir is set to the location of Qt\n"
+	" 2) %QTDIR% is set to the location of Qt\n"
+	"\n"
+	"This program will use the first one it finds to be true, checked in the above order.\n"
+	"\n";
+
 class ConfUsageOpt
 {
 public:
@@ -353,6 +456,62 @@ public:
 		return str.toLatin1();
 	}
 
+	QByteArray generateExe()
+	{
+		// main options
+		mainopts.clear();
+		mainopts += ConfOpt("qtdir", "path", "EX_QTDIR", "Directory where Qt is installed.");
+
+		if(libmode)
+			mainopts += ConfOpt("static", QString(), "QC_STATIC", "Create a static library instead of shared.");
+
+		// combine main and extra opts together
+		all = mainopts + appopts + depopts;
+
+		QByteArray out = get_configexe_stub();
+		QByteArray sig = "QCONF_CONFIGWIN_BLOCKSIG_68b7e7d7";
+		QByteArray datasec = makeDatasec();
+		out += sig;
+		out += lenval(datasec);
+		return out;
+	}
+
+	QByteArray makeDatasec() const
+	{
+		QByteArray out;
+		QByteArray buf(4, 0);
+
+		out += lenval(genUsageOutput().toLatin1());
+
+		write32((quint8 *)buf.data(), all.count());
+		out += buf;
+		for(int n = 0; n < all.count(); ++n)
+		{
+			const ConfOpt &i = all[n];
+			out += lenval(i.name.toLatin1());
+			out += lenval(i.var.toLatin1());
+			if(!i.arg.isEmpty())
+				out += (char)0;
+			else
+				out += (char)1;
+		}
+
+		write32((quint8 *)buf.data(), 5);
+		out += buf;
+		out += embed_file("modules.cpp", filter_backslashes(filemodulescpp));
+		out += embed_file("modules_new.cpp", filter_backslashes(filemodulesnewcpp));
+		out += embed_file("conf4.h", filter_backslashes(fileconfh));
+		out += embed_file("conf4.cpp", filter_backslashes(fileconfcpp));
+		out += embed_file("conf4.pro", filter_backslashes(fileconfpro));
+
+		out += lenval(name.toLatin1());
+		out += lenval(profile.toLatin1());
+
+		out += lenval(formatBlock(qt4_info_str_win).toLatin1());
+
+		return out;
+	}
+
 private:
 	QString genHeader()
 	{
@@ -382,7 +541,7 @@ private:
 		return str;
 	}
 
-	int getUsageIndent(const QList<ConfUsageOpt> list)
+	int getUsageIndent(const QList<ConfUsageOpt> list) const
 	{
 		int largest = 0;
 		for(QList<ConfUsageOpt>::ConstIterator it = list.begin(); it != list.end(); ++it) {
@@ -394,7 +553,7 @@ private:
 		return largest;
 	}
 
-	QList<ConfUsageOpt> optsToUsage(const QList<ConfOpt> &list)
+	QList<ConfUsageOpt> optsToUsage(const QList<ConfOpt> &list) const
 	{
 		QList<ConfUsageOpt> out;
 		for(QList<ConfOpt>::ConstIterator it = list.begin(); it != list.end(); ++it) {
@@ -404,7 +563,7 @@ private:
 		return out;
 	}
 
-	QString genUsageSection(const QString &title, const QList<ConfUsageOpt> &list)
+	QString genUsageSection(const QString &title, const QList<ConfUsageOpt> &list) const
 	{
 		QString str;
 		str += title;
@@ -444,6 +603,30 @@ private:
 		return str;
 	}
 
+	QString genUsageOutput() const
+	{
+		QString str =
+		"Usage: $0 [OPTION]...\n\n"
+		"This script creates necessary configuration files to build/install.\n\n";
+
+		QList<ConfUsageOpt> list = optsToUsage(mainopts);
+		list += ConfUsageOpt("verbose",  "",   "Show extra configure output.");
+		list += ConfUsageOpt("help",   "",     "This help text.");
+		str += genUsageSection("Main options:", list);
+
+		if(!appopts.isEmpty()) {
+			list = optsToUsage(appopts);
+			str += genUsageSection("Project options:", list);
+		}
+
+		if(!depopts.isEmpty()) {
+			list = optsToUsage(depopts);
+			str += genUsageSection("Dependency options:", list);
+		}
+
+		return str;
+	}
+
 	QString genFindStuff()
 	{
 		QString str;
@@ -694,10 +877,11 @@ private:
 		"# run qmake -v and check version\n"
 		"qmake_check_v4() {\n"
 		"	if [ -x \"$1\" ]; then\n"
-		"		if echo `$1 -v 2>&1` | grep \"Qt version 4\\.\" >/dev/null 2>&1; then\n"
-		"			return 0\n"
-		"		elif [ \"$QC_VERBOSE\" = \"Y\" ]; then\n"
-		"			echo \"Warning: $1 not for Qt 4\"\n"
+		"       local v=`$1 -query QT_VERSION 2>&1`\n"
+		"       vmaj=\"${v%%.*}\"\n"
+		"       case \"${v}\" in ?.?.?) [ \"$vmaj\" = 4 -o \"$vmaj\" = 5 ] && return 0 ;; esac\n"
+		"		if [ \"$QC_VERBOSE\" = \"Y\" ]; then\n"
+		"			echo \"Warning: $1 not for Qt 4 or 5\"\n"
 		"		fi\n"
 		"	fi\n"
 		"	return 1\n"
@@ -710,7 +894,7 @@ private:
 		"fi\n"
 		"\n"
 		"qm=\"\"\n"
-		"names=\"qmake-qt4 qmake4 qmake\"\n"
+		"names=\"qmake-qt5 qmake5 qmake-qt4 qmake4 qmake\"\n"
 		"\n"
 		"# qt4 check: --qtdir\n"
 		"if [ -z \"$qm\" ] && [ ! -z \"$EX_QTDIR\" ]; then\n"
@@ -790,6 +974,37 @@ private:
 		"if [ \"$QC_VERBOSE\" = \"Y\" ]; then\n"
 		"	echo qmake found in $qm\n"
 		"fi\n\n";
+
+		str +=
+		"# try to determine the active makespec\n"
+		"defmakespec=$QMAKESPEC\n"
+		"if [ -z \"$defmakespec\" ]; then\n"
+		"	if $WHICH readlink >/dev/null 2>&1; then\n"
+		"		READLINK=`$WHICH readlink`\n"
+		"	fi\n"
+		"	if [ ! -z \"$READLINK\" ]; then\n"
+		"		qt_mkspecsdir=`$qm -query QT_INSTALL_DATA`/mkspecs\n"
+		"		if [ -d \"$qt_mkspecsdir\" ] && [ -h \"$qt_mkspecsdir/default\" ]; then\n"
+		"			defmakespec=`$READLINK $qt_mkspecsdir/default`\n"
+		"		fi\n"
+		"	fi\n"
+		"fi\n"
+		"\n"
+		"if [ \"$QC_VERBOSE\" = \"Y\" ]; then\n"
+		"	echo makespec is $defmakespec\n"
+		"fi\n"
+		"\n"
+		"qm_spec=\"\"\n"
+		"# if the makespec is macx-xcode, force macx-g++\n"
+		"if [ \"$defmakespec\" = \"macx-xcode\" ]; then\n"
+		"	qm_spec=macx-g++\n"
+		"	QMAKESPEC=$qm_spec\n"
+		"	export QMAKESPEC\n"
+		"	if [ \"$QC_VERBOSE\" = \"Y\" ]; then\n"
+		"		echo overriding makespec to $qm_spec\n"
+		"	fi\n"
+		"fi\n\n";
+
 		return str;
 	}
 
@@ -810,7 +1025,11 @@ private:
 
 		if(qt4) {
 			str +=
-			"	$qm conf4.pro >/dev/null\n"
+			"	if [ ! -z \"$qm_spec\" ]; then\n"
+			"		$qm -spec $qm_spec conf4.pro >/dev/null\n"
+			"	else\n"
+			"		$qm conf4.pro >/dev/null\n"
+			"	fi\n"
 			"	$MAKE clean >/dev/null 2>&1\n"
 			"	$MAKE >../conf.log 2>&1\n";
 		}
@@ -880,6 +1099,8 @@ private:
 			str += QString("export QC_PROFILE\n");
 			str += QString("QC_QMAKE=$qm\n");
 			str += QString("export QC_QMAKE\n");
+			str += QString("QC_QMAKESPEC=$qm_spec\n");
+			str += QString("export QC_QMAKESPEC\n");
 			str += QString("QC_MAKETOOL=$MAKE\n");
 			str += QString("export QC_MAKETOOL\n");
 		}
@@ -1080,11 +1301,6 @@ public:
 	}
 };
 
-const char * str2out(const QString &str)
-{
-	return qPrintable(str);
-}
-
 enum VersionMode { VersionMin, VersionExact, VersionMax, VersionAny };
 
 class Dep
@@ -1118,6 +1334,7 @@ public:
 		nobindir = false;
 		useincdir = false;
 		uselibdir = false;
+		usedatadir = false;
 		qt4 = false;
 		byoq = false;
 	}
@@ -1242,7 +1459,7 @@ int main(int argc, char **argv)
 			fname = conf.name + ".qc";
 			QFile f(fname);
 			if(!f.open(QFile::WriteOnly | QFile::Truncate)) {
-				printf("qconf: unable to write %s\n", str2out(fname));
+				printf("qconf: unable to write %s\n", qPrintable(fname));
 				return 1;
 			}
 			QDomDocument doc;
@@ -1290,19 +1507,19 @@ int main(int argc, char **argv)
 	if(!skipLoad) {
 		QFile f(fname);
 		if(!f.open(QFile::ReadOnly)) {
-			printf("qconf: error reading %s\n", str2out(f.fileName()));
+			printf("qconf: error reading %s\n", qPrintable(f.fileName()));
 			return 1;
 		}
 		QDomDocument doc;
 		if(!doc.setContent(&f)) {
-			printf("qconf: error parsing %s\n", str2out(f.fileName()));
+			printf("qconf: error parsing %s\n", qPrintable(f.fileName()));
 			return 1;
 		}
 		f.close();
 
 		QDomElement base = doc.documentElement();
 		if(base.tagName() != "qconf") {
-			printf("qconf: bad format of %s\n", str2out(f.fileName()));
+			printf("qconf: bad format of %s\n", qPrintable(f.fileName()));
 			return 1;
 		}
 		conf = xmlToConf(base);
@@ -1324,13 +1541,18 @@ int main(int argc, char **argv)
 		confdirpath = localDataPath + "/conf";
 	}
 	else {
+#ifdef DATADIR
 		moddirs += QString(DATADIR) + "/qconf/modules";
 		confdirpath = QString(DATADIR) + "/qconf/conf";
+#else
+		moddirs += "./modules";
+		confdirpath = "./conf";
+#endif
 	}
 
 	QDir confdir(confdirpath);
 	if(!confdir.exists()) {
-		printf("qconf: %s does not exist.\n", str2out(confdir.path()));
+		printf("qconf: %s does not exist.\n", qPrintable(confdir.path()));
 		return 1;
 	}
 
@@ -1339,7 +1561,7 @@ int main(int argc, char **argv)
 	if(conf.qt4) {
 		f.setFileName(confdir.filePath("conf4.h"));
 		if(!f.open(QFile::ReadOnly)) {
-			printf("qconf: cannot read %s\n", str2out(f.fileName()));
+			printf("qconf: cannot read %s\n", qPrintable(f.fileName()));
 			return 1;
 		}
 		confh = f.readAll();
@@ -1348,7 +1570,7 @@ int main(int argc, char **argv)
 
 	f.setFileName(confdir.filePath(conf.qt4 ? "conf4.cpp" : "conf.cpp"));
 	if(!f.open(QFile::ReadOnly)) {
-		printf("qconf: cannot read %s\n", str2out(f.fileName()));
+		printf("qconf: cannot read %s\n", qPrintable(f.fileName()));
 		return 1;
 	}
 	QByteArray confcpp = f.readAll();
@@ -1356,15 +1578,15 @@ int main(int argc, char **argv)
 
 	f.setFileName(confdir.filePath(conf.qt4 ? "conf4.pro" : "conf.pro"));
 	if(!f.open(QFile::ReadOnly)) {
-		printf("qconf: cannot read %s\n", str2out(f.fileName()));
+		printf("qconf: cannot read %s\n", qPrintable(f.fileName()));
 		return 1;
 	}
 	QByteArray confpro = f.readAll();
 	f.close();
 	confpro += "\nDEFINES += HAVE_MODULES\n";
 
-	printf("Project name: %s\n", str2out(conf.name));
-	printf("Profile: %s\n", str2out(conf.profile));
+	printf("Project name: %s\n", qPrintable(conf.name));
+	printf("Profile: %s\n", qPrintable(conf.profile));
 	printf("Deps: ");
 	if(conf.deps.isEmpty())
 		printf("none\n");
@@ -1372,7 +1594,7 @@ int main(int argc, char **argv)
 		bool first = true;
 		for(QList<Dep>::ConstIterator it = conf.deps.begin(); it != conf.deps.end(); ++it) {
 			const Dep &dep = *it;
-			printf("%s%s%s", first ? "": " ", str2out(dep.name), dep.required ? "*": "");
+			printf("%s%s%s", first ? "": " ", qPrintable(dep.name), dep.required ? "*": "");
 			first = false;
 		}
 		printf("\n");
@@ -1426,7 +1648,7 @@ int main(int argc, char **argv)
 			return 1;
 		}
 		QFile mod(fi.filePath());
-		if(!mod.open(QFile::ReadOnly)) {
+		if(!mod.open(QFile::ReadOnly | QFile::Text)) {
 			printf("qconf: error opening '%s'!\n", qPrintable(mod.fileName()));
 			return 1;
 		}
@@ -1517,9 +1739,26 @@ int main(int argc, char **argv)
 	QByteArray cs = cg.generate();
 	out.write(cs);
 	out.close();
+#ifdef Q_OS_UNIX
 	chmod("configure", 0755);
+#endif
 
 	printf("'configure' written.\n");
 
+	if(conf.qt4)
+	{
+		// write configexe
+		out.setFileName("configure.exe");
+		if(!out.open(QFile::WriteOnly | QFile::Truncate)) {
+			printf("qconf: error writing configure.exe\n");
+			return 1;
+		}
+		cs = cg.generateExe();
+		out.write(cs);
+		out.close();
+
+		printf("'configure.exe' written.\n");
+	}
+
 	return 0;
 }
